/*

  (C) 1998 Marco Falcioni UCLA

  Cell methods

*/
/* includes */
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <limits.h>

/* from libgjm-1.3 */
#include "constants.h"
#include "matrices.h" 
#include "memory.h"
#include "error.h"

#include "atoms.h"
#include "cell.h"
#include "misc.h"
#include "verbose.h"

/* 
   allocate and polish new Cell
   ANGLES IN RADIANS
*/

struct Cell * new_Cell(const double a, const double b, const double c,
		       const double alpha, const double beta,
		       const double gamma)
{
  struct Cell * C;
  long i,j,k;
  unsigned int m;
  double ca,cb,cg,sg,A,B;
  double twopi = 2.0 * constPi_;

  if(a <= 0.0 || b <= 0.0 || c <= 0.0 || 
     alpha <= 0.0 || beta <= 0.0 || gamma <= 0.0) {
    push_generic_error("new_Cell","non-positive args passed");
    exit_clean();
  }

  if(alpha > twopi || beta > twopi || gamma > twopi) {
    push_generic_error("new_Cell","angles greater than 2 Pi... why?");
  }

  C = (struct Cell *) Malloc(sizeof(struct Cell));

  if(C == 0) {
    push_error("new_Cell",Error_Alloc_);
    exit_clean();
  }

  C->csym = 0; /* default to non centrosymmetric */
  
  C->a = a;
  C->b = b;
  C->c = c;
  C->alpha = fmod(alpha,twopi);
  C->beta  = fmod(beta,twopi);
  C->gamma = fmod(gamma,twopi); 

  C->abc_to_xyz = new_matrix_d(3);
  C->xyz_to_abc = new_matrix_d(3);

  ca = cos(alpha);
  cb = cos(beta);
  cg = cos(gamma);
  sg = sin(gamma);

  C->cosalpha = ca;
  C->cosbeta = cb;
  C->cosgamma = cg;
  
  A = (ca - cb*cg)/sg;
  B = sqrt( 1.0 - cb*cb - A*A ); 

  /* how to translate abc components in to spatial positions */

  C->abc_to_xyz[0][0] = a;
  C->abc_to_xyz[0][1] = b * cg;
  C->abc_to_xyz[0][2] = c * cb;
  C->abc_to_xyz[1][0] = 0.0;
  C->abc_to_xyz[1][1] = b * sg;
  C->abc_to_xyz[1][2] = c * A;
  C->abc_to_xyz[2][0] = 0.0;
  C->abc_to_xyz[2][1] = 0.0;
  C->abc_to_xyz[2][2] = c * B;

  /* inverse of above matrix: from xyz to abc */

  C->xyz_to_abc[0][0] = 1.0 / a;
  C->xyz_to_abc[0][1] = - cg / ( a * sg);
  C->xyz_to_abc[0][2] = (cg * A - sg * cb)/(sg * a * B);
  C->xyz_to_abc[1][0] = 0.0;
  C->xyz_to_abc[1][1] = 1.0/(b * sg);
  C->xyz_to_abc[1][2] = -A/(b * sg * B);
  C->xyz_to_abc[2][0] = 0.0;
  C->xyz_to_abc[2][1] = 0.0;
  C->xyz_to_abc[2][2] = 1.0/(c * B);

  /* reciprocal lattice vectors */

  C->rec_vec = C->xyz_to_abc; /* they seem to look like the same matrix...*/

  /* set up 3^3 - 1 translation vectors */
  C->transl = new_vector_dp(26);

  m = 0;
  for(i = -1; i < 2; i++) {
    for(j = -1; j < 2; j++) {
      for(k = -1; k < 2; k++) {
	if(k == 0 && j == 0 && i == 0) continue;
	C->transl[m] = new_vector_d(3);
	C->transl[m][0] = (double) i;
	C->transl[m][1] = (double) j;
      	C->transl[m][2] = (double) k;
	m++;
      }
    }
  }

  /* here we find the parallelepiped bounding the unit cell */

  C->lnl[0] = 10000.0;
  C->lnl[1] = 10000.0;
  C->lnl[2] = 10000.0;

  C->ufr[0] = -10000.0;
  C->ufr[1] = -10000.0;
  C->ufr[2] = -10000.0;


  for(i = 0; i < 2; i++) {
    for(j = 0; j < 2; j++) {
      for(k = 0; k < 2; k++) {
	int n;
	double a[3], b[3];

	a[0] = (double) i;
	a[1] = (double) j;
	a[2] = (double) k;

	matvec_multiply_d(b,C->abc_to_xyz,a,3);
	for(n=0;n<3;n++) {
	  C->lnl[n] = min(C->lnl[n], b[n]);
	  C->ufr[n] = max(C->ufr[n],b[n]);
	}
      }
    }
  }

  return C;
}

void delete_Cell (struct Cell * C)
{
  int i;
  if(C != 0) {
    free_matrix_d(C->abc_to_xyz,3);
    free_matrix_d(C->xyz_to_abc,3);
    
    for(i=0; i<27;i++)
      free_vector(C->transl[i]);
    
    Free(C->transl);
    
    Free(C);
  }
  return;
}

/* Reads and allocates cell from stream. Comments preceded by # */
struct Cell * read_Cell(FILE * stream)
{
  struct Cell * C;
  char line[1024];
  double a[2][3];
  int i = 0;
  char * p;

  if(stream == 0) {
    push_generic_error("read_Cell","(nil) arg passed");
    exit_clean();
  }

  push_verbose_message("read_Cell","reading cell data");

  /* only need to read six numbers! */

  do {
    p = get_line(line, sizeof(line), stream);
    if(p == 0 ||  sscanf(p,"%lf %lf %lf",&a[i][0],&a[i][1],&a[i][2]) != 3) {
      push_generic_error("read_Cell","invalid line for cell data");
      exit_clean();
    }
    i++;
  } while ( i < 2 );

  if(a[0][0] < 0.0   || a[0][1] < 0.0   || a[0][2] < 0.0 ||
     a[1][0] < 0.0   || a[1][1] < 0.0   || a[1][2] < 0.0 ||
     a[1][0] > 180.0 || a[1][1] > 180.0 || a[1][2] > 180.0 ) {
    push_generic_error("read_Cell","invalid cell data");
    exit_clean();
  }

   /* convert to radians */
  a[1][0] *= (constPi_/180.0) ;
  a[1][1] *= (constPi_/180.0) ;
  a[1][2] *= (constPi_/180.0) ;

  C = new_Cell(a[0][0], a[0][1], a[0][2], a[1][0], a[1][1], a[1][2]);

  if(!(p = get_line(line, sizeof(line), stream))) {
    push_generic_error("read_Cell", "invalid cell data");
    exit_clean();
  }

  p[strlen(p) -1] = '\0'; /* remove newline */

  if(strcmp(p,"CENTROSYMMETRIC") == 0) C->csym = 1;

  return C;
}

struct Cell * read_Cell_XTL(FILE * stream)
{
  struct Cell * C;
  char line[1024];
  double a[6];
  char * p;

  if(stream == 0) {
    push_generic_error("read_Cell_XTL","(nil) arg passed");
    exit_clean();
  }

  push_verbose_message("read_Cell_XTL","reading cell data");

  /* read the CELL line */
  p = get_line(line, sizeof(line), stream); 

  if(!strcmp(p,"CELL")){
    push_generic_error("read_Cell_XTL","cannot seem to find a cell");
    exit_clean();
  }

  /* only need to read six numbers! */
  p = get_line(line, sizeof(line), stream);
  if(p == 0 ||  sscanf(p,"%lf %lf %lf %lf %lf %lf",
		       &a[0],&a[1],&a[2],&a[3],&a[4],&a[5]) != 6) {
    push_generic_error("read_Cell","invalid line for cell data");
    exit_clean();
  }

  if(a[0] < 0.0   || a[1] < 0.0   || a[2] < 0.0 ||
     a[3] < 0.0   || a[4] < 0.0   || a[5] < 0.0 ||
     a[3] > 180.0 || a[4] > 180.0 || a[5] > 180.0 ) {
    push_generic_error("read_Cell","invalid cell data");
    exit_clean();
  }

   /* convert to radians */
  a[3] *= (constPi_/180.0) ;
  a[4] *= (constPi_/180.0) ;
  a[5] *= (constPi_/180.0) ;

  C = new_Cell(a[0], a[1], a[2], a[3], a[4], a[5]);

  return C;
}

void write_Cell (const struct Cell * C, FILE * stream)
{
  if(C == 0 || stream == 0) {
    push_generic_error("write_Cell","(nil) args passed");
    exit_clean();
  }
  fprintf(stream, "# a b c, alpha, beta, gamma\n");
  fprintf(stream, "%6.3f %6.3f %6.3f\n", C->a, C->b, C->c);
  fprintf(stream, "%6.3f %6.3f %6.3f\n", 180.0 * C->alpha / constPi_, 
	  180.0 * C->beta / constPi_, 180.0 * C->gamma / constPi_);
  if(C->csym == 1)  fprintf(stream, "CENTROSYMMETRIC\n") ;
  else fprintf(stream, "NONCENTROSYMMETRIC\n") ;
  fprintf(stream, "# lower near left corner\n");
  fprintf(stream, "# %f %f %f\n",C->lnl[0], C->lnl[1], C->lnl[2]);
  fprintf(stream, "#  upper far right corner\n");
  fprintf(stream, "# %f %f %f\n",C->ufr[0], C->ufr[1], C->ufr[2]);
  return;
}


void write_Cell_XTL (const struct Cell * C, FILE * stream)
{
  if(C == 0 || stream == 0) {
    push_generic_error("write_Cell","(nil) args passed");
    exit_clean();
  }
  fprintf(stream, "CELL\n  ");
  fprintf(stream, "%6.3f    %6.3f    %6.3f   %9.5f   %9.5f   %9.5f\n", 
	  C->a, C->b, C->c, 180.0 * C->alpha / constPi_, 
	  180.0 * C->beta / constPi_, 180.0 * C->gamma / constPi_);
  return;
}

/* Returns the volume of the cell */
double volume_Cell (const struct Cell * C)
{
  double volume;
  double ca, cb, cg; 

  if(C == 0) {
    push_generic_error("volume_Cell","(nil)arg passed");
    exit_clean();
  }

  ca = C->cosalpha;
  cb = C->cosbeta;
  cg = C->cosgamma;
  
  volume = C->a * C->b * C->c *
    sqrt(1.0 - ca * ca - cb * cb - cg * cg + 2.0 * ca * cb * cg);

  return volume;
}


/* 
   Returns the distance between two cell points in real space units 
*/
 
double dist2_xyz(const struct Cell * C, 
		struct Atom * A, struct Atom * B)
{
  double dist2;
  double x,y,z,a,b,c,ca,cb,cg;
#ifndef NOCHECKS
  if(A == 0 || B == 0 || C == 0) {
    push_generic_error("dist_xyz","(nil) args passed");
    exit_clean();
  }
#endif
  ca = C->cosalpha;
  cb = C->cosbeta;
  cg = C->cosgamma;
  
  a = C->a;
  b = C->b;
  c = C->c;

  x = A->ra - B->ra;
  y = A->rb - B->rb;
  z = A->rc - B->rc;

  dist2 = x * x * a * a  +  y * y *  b * b  +  z * z * c * c + 
    2.0*(x * y * a * b * cg  +  y * z * b * c * ca  +  z * x * c * a * cb); 

  return dist2;
}

double min_dist2_xyz(const struct Cell * C, 
		const struct Atom * A, const struct Atom * B)
{
  double dist2;
  double a, b, c;
  double ca, cb, cg;
  double x, y, z;

  if(A == 0 || B == 0 || C == 0) {
    push_generic_error("dist_xyz","(nil) args passed");
    exit_clean();
  }

  ca = C->cosalpha;
  cb = C->cosbeta;
  cg = C->cosgamma;
  
  a = C->a;
  b = C->b;
  c = C->c;

  x = A->ra - B->ra;
  y = A->rb - B->rb;
  z = A->rc - B->rc;

  if(x > 0.5) x -= 1.0;
  else if (x < -0.5) x += 1.0;

  if(y > 0.5) y -= 1.0;
  else if (y < -0.5) y += 1.0;

  if(z > 0.5) z -= 1.0;
  else if (z < -0.5) z += 1.0;

  dist2 = x * x * a * a  +  y * y *  b * b  +  z * z * c * c + 
    2.0*(x * y * a * b * cg  +  y * z * b * c * ca  +  z * x * c * a * cb);
  
  return dist2;
}

/* 
   return the distance between two planes form the miller indeces specified
   cfr. D.E. Sands Introduction to Crystallography page 68  interplanar 
   spacing given a set of generalized miller indeces 
*/
double dist2_plane (const struct Cell * C, 
		   const long _h, const long _k, const long _l)
{
  double dist;
  double a, b, c;
  double a2, b2, c2;
  double ca, cb, cg;
  double sa, sb, sg;
  double V;
  double h2, k2, l2;

  double H = (double) _h, 
    K = (double) _k, 
    L = (double) _l;


  if(C == 0) {
    push_generic_error("dist_plane","(nil) arg passed");
    exit_clean();
  }

  ca = C->cosalpha;
  cb = C->cosbeta;
  cg = C->cosgamma;

  sa = sin(C->alpha);
  sb = sin(C->beta);
  sg = sin(C->gamma);

  a = C->a;
  b = C->b;
  c = C->c;
  
  a2 = a*a;
  b2 = b*b;
  c2 = c*c;

  h2 = (double) (_h * _h);
  k2 = (double) (_k * _k);
  l2 = (double) (_l * _l);

  V = volume_Cell(C);

  dist = (h2 * b2 * a2 * sa * sa + 
	      k2 * a2 * c2 * sb * sb +
	      l2 * a2 * b2 * sg * sg + 
	      2.0 * ( H * L * a  * b2 * c  * (ca * cg - cb) + 
		      H * K * a  * b  * c2 * (ca * cb - cg) + 
		      K * L * a2 * b  * c  * (cb * cg - ca)));
	 
  dist = V*V/dist;

  return dist;
}







