/*
  (C) 1997 Marco Falcioni

  creation destruction and all of that for grids

*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "error.h"
#include "memory.h"
#include "matrices.h"

#include "atoms.h"
#include "cell.h"
#include "positions.h"
#include "link.h"
#include "box.h"
#include "grid.h"
#include "params.h"
#include "misc.h"

#define EPS 1e-10

struct Grid * new_Grid(const struct Params * P, const struct Cell * C, unsigned int radius)
{
  double bond;
  double sizes[3];
  unsigned int i, Lx, Ly, Lz;
  struct Grid * grid;

  if(C == 0){
    push_generic_error("new_Grid","(nil) arg passed");
    exit_clean();
  }

  grid = (struct Grid *) Malloc(sizeof(struct Grid));

  /* the grid spacing. It must be a little bigger than the maximum
     range of interactions */

  bond = P->frange; 
  
  grid->spacing = bond;
  grid->inv_spacing = 1.0/bond;
  grid->filled = 0;
  grid->n_boxes = 1;

  /* compute center of box */
  sizes[0] = 0.5;
  sizes[1] = 0.5;
  sizes[2] = 0.5;
    
  matvec_multiply_d(grid->center, C->abc_to_xyz, sizes, 3);
    
  if (radius == 0) { /* this defaults to smallest possible grid */
    
    /* compute the number of boxes in the grid */
    
    for(i=0; i<3; i++) {
      sizes[i] = C->ufr[i] - C->lnl[i] + 2.0 * bond;
      grid->sizes[i] = (unsigned int) ceil(sizes[i]/bond);
      grid->n_boxes *= grid->sizes[i];    /* volume */
      grid->corner[i] = C->lnl[i] - bond; /* real space position of bottommost
					     point */
      grid->cells[i] = 1;
    }
    
    if(grid->n_boxes > 2048) { /* 40^3 cubic Angstrom! if bond = 5 */
      char buf[1024];
      sprintf(buf,"way too many boxes...%d",grid->n_boxes);
      push_generic_error("new_Grid",buf);
      exit_clean();
    }
  } else {
    /* we specify a box size and we go with it! */
    for(i=0; i<3; i++) {
      unsigned int a;
      sizes[i] = C->ufr[i] - C->lnl[i];
      a = radius * (unsigned int) ceil(sizes[i]/bond);
      grid->cells[i] = radius;
      grid->sizes[i] = (2 * radius + 1) * a;
      grid->n_boxes *= grid->sizes[i];
      grid->corner[i] = C->lnl[i] - (double) radius * a - bond;
    }
                                                                         ;
    if(radius > 10) { /* 40^3 boxes...*/
      char buf[1024];
      sprintf(buf,"way too many boxes...%d",grid->n_boxes);
      push_generic_error("new_Grid",buf);
      exit_clean();
    }
    
  }

  Lx = grid->sizes[0];
  Ly = grid->sizes[1];
  Lz = grid->sizes[2];
  
  grid->box_list = 
    (struct Box **) Malloc(grid->n_boxes * sizeof(struct Box *));

  for(i = 0; i < grid->n_boxes; i++) {
    grid->box_list[i] = new_Box();
  }

  for(i = 0; i < Lz; i++) {
    unsigned int j;
    for(j = 0; j < Ly; j++) {
      unsigned int k;
      for(k = 0; k < Lx; k++) {
	long n,b=0;
	unsigned int s;
	struct Box * box;

	s = ( i * Ly + j ) * Lx + k;
	box = grid->box_list[s];
	/* loop over 26 neighbors */ 
	for(n = -1; n < 2; n++) {
	  long m, nn = (long) i+n;
	  for(m = -1; m < 2; m++) {
	    long p, mm = (long) j+m;
	    for(p = -1; p < 2; p++) {
	      long pp = (long) k+p;
	      long r;
	      if(pp >= 0 && pp < (long) Lx && mm >= 0 && 
		 mm < (long) Ly && nn >= 0 && nn < (long) Lz) {
		r = (nn * Ly + mm) * Lx + pp;
		if(r != s) { /* only if it is not same box */
		  box->near[b] = grid->box_list[r];
		  b++;
		}
	      }
	    }
	  }
	}
	box->n = b;
      }
    }
  }

  return grid;  

}

void delete_Grid(struct Grid * G)
{
  unsigned long i;
  if(G == 0) return;
  for(i = 0; i < G->n_boxes; i++) {
    delete_Box(G->box_list[i]);   
  }
  Free(G->box_list);
  Free(G);
  return;
}


void untag_Grid(struct Grid * G)
{
  unsigned long i;
  if (G == 0) return;
  for(i = 0; i < G->n_boxes; i++) {
    struct Link * me;
    for(me = G->box_list[i]->list->head; me != 0; me = me->next){
      me->atom->tag = 0;
    }      
  }
  return;
}

void clean_Grid(struct Grid * G)
{
  unsigned long i;

  if(G == 0) {
    push_generic_error("clean_Grid","(nil) arg passed");
    exit_clean();
  }

  if(G->filled == 0) return;

  for(i = 0; i < G->n_boxes; i++) {
    clean_Chain(G->box_list[i]->list);   
  }

  G->filled = 0;

  return;
}

void print_Grid(const struct Grid * grid, FILE * stream) 
{
  unsigned int i;
  if(grid == 0 || stream == 0) {
    push_generic_error("print_Grid","(nil) args passed");
    exit_clean();
  }
  fprintf(stream,"Grid at %p\n", grid);
  fprintf(stream,"  G->n_boxes = %d", grid->n_boxes);
  fprintf(stream,"  G->sizes[3] = {%d, %d, %d}\n", 
	  grid->sizes[0],grid->sizes[1],grid->sizes[2]);
  fprintf(stream,"  G->box_list at %p {\n", grid->box_list);
  
  for(i = 0; i < grid->n_boxes; i++) {
    fprintf(stream,"  G->box_list[%d]\n", i);
    print_Box(grid->box_list[i], stream);
  }
  fprintf(stream,"  } end Grid\n");
  return;
}


struct Box * index_atom_Grid (const struct Grid * G, 
			      struct Atom * A,
			      const struct Cell * C, long * ii) 
{
  double pos[3];
  long i, j[3];

#ifndef NOCHECKS
  if(G == 0 || pos == 0 || C == 0) {
    push_generic_error("index_atom_Grid","(nil) args passed");
    exit_clean();
  }
#endif

  if(ii == 0) {
    ii = j;
  }

  xyz_Atom(A,C);

  pos[0] = A->rx;
  pos[1] = A->ry;
  pos[2] = A->rz;

  for(i = 0; i < 3; i++) {
    double dist = pos[i] - G->corner[i];
    if(dist < 0.0) { /* atom falls outside grid */
      return 0;
    } 

    dist *= G->inv_spacing;
    ii[i] = (long) floor(dist);
    if(ii[i] >= (long) G->sizes[i]) {
      /* atom falls outside grid also */
      return 0;
    } 
  }

  i = (ii[2] * (long)G->sizes[1] + ii[1]) * (long) G->sizes[0] + ii[0];

  return G->box_list[i];

}

void fill_Grid(struct Grid * G, 
	       struct Positions * T    , 
	       const struct Cell * C)
{
  struct Link * Ld;
  struct Atom * A1;

  if(T == 0 || C == 0 || G == 0){
    push_generic_error("fill_Grid","(nil) args passed");
    exit_clean();
  }

  if(G->filled == 1) {
    push_generic_error("fill_Grid","trying to fill a filled grid");
    exit_clean();
  }

  if(T->merged != 1) {
    push_generic_error("fill_Grid","merge the mesh first");
    exit_clean();
  }

  /* radius of the ball containing the unit cell we'll fill
     with atoms */

  A1 = new_Atom();

  G->filled = 1;

  for(Ld = T->distinct->head; Ld != 0; Ld = Ld->next) {
    long m;
    double a[3];
    struct Link * Lb;
    struct Box * box;
    struct Atom * A;
    /* add the unit cell atoms in the gird */

    mod_Atom(Ld->atom);

    box = index_atom_Grid(G, Ld->atom, C, 0);
    if(box == 0) {
      push_generic_error("fill_Grid", "unit cell atom not in grid?");
      exit_clean();
    }
    add_Link(new_Link(Ld->atom), box->list);

    for(m = - (long) G->cells[0]; m <= (long) G->cells[0]; m++) {
      long n;
      for(n = - (long) G->cells[1]; n <= (long) G->cells[1]; n++) {
	long p;
	for(p = - (long) G->cells[2]; p <= (long) G->cells[2]; p++) {

	  if(m == 0 && n == 0 && p == 0) continue;

	  a[0] = Ld->atom->ra + (double) m;
	  a[1] = Ld->atom->rb + (double) n;
	  a[2] = Ld->atom->rc + (double) p;

	  move_Atom(A1,a);
	  
	  box = index_atom_Grid(G, A1, C, 0);
	  
	  if(box != 0) {
	    A = new_Atom();
	    move_Atom(A, a);
	    clone_Atom(A, Ld->atom);
	    Lb = new_Link(A);
	    add_Link(Lb, box->list);
	  }
	}
      }
    }
  } /* end loop over distinct atoms */

  delete_Atom(A1);

  return;
}
