/*

  (C) 1998 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html

   1998 (C) Marco Falcioni UCLA

   this routine proposes a new configuration
   then accepts the new configuration according to a Metropolis
   test.

*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <time.h>

#include "error.h"

#include "atoms.h"
#include "link.h"
#include "params.h"
#include "positions.h"
#include "atoms.h"
#include "cell.h"
#include "grid.h"
#include "functions.h"
#include "chain.h"
#include "misc.h"
#include "trial.h"
#include "symmetry.h"
#include "rad_pattern.h"
#include "oogl.h"
#include "conf.h"
#include "molecule.h"
#include "vector_functions.h"

#undef TRIALS

#undef MC_ENERGY

double monte_carlo (const struct Params * P, const struct Cell * C,
		    const struct Sym_group * G, struct Rad_pattern * RP,
		    struct Positions * T1, struct Positions * T2,
		    struct Grid * Gr, struct Conf * Min) 
{
  double ret = 0.0;
  double Beta = 1.0 / P->temperature;
  unsigned long iter, looper;
  unsigned accepted = 0; 
  double o_energy= 0.0, n_energy = 0.0;
  double o_RP_distance = 0.0, n_RP_distance = 0.0;
  struct Trial * Delta;
  unsigned long Accepted = 0, Tried = 0;
  struct Conf * c1, *c2;

#ifdef MC_ENERGY
  FILE * AA, *BB = 0;
  AA = safe_open("energies","w+");
  if(T1->Nmol) BB = safe_open("molecule","w+");
  run_ID(AA);
  if(BB) run_ID(BB);
#endif

#ifndef NOCHECKS
  if(P == 0 || C == 0 || G == 0 || T1 == 0 || T2 == 0 || Gr == 0) {
    push_generic_error("monte_carlo", "(nil) args passed");
    exit_clean();
  }
#endif
  c1 = new_Conf(T1->n_unique);
  c2 = new_Conf(T1->n_unique);

  Delta = new_Trial(P->radius);

  copy_Positions(T1,T2);
  symmetrize_Positions(T2,G);
  mod_Positions(T2);   

  merge_positions(T2,P,C);

  clean_Grid(Gr);
  fill_Grid(Gr,T2,C);
  o_energy = energy (P, C, T2, Gr, c1);
  
  if(RP != 0) {
    get_intensity(RP, T2, C, G);
    scale_Rad_pattern(RP);
    o_RP_distance = distance_Rad_pattern(RP);
  }
  

  c1->RP_energy = o_RP_distance;
  c1->tot_energy =  P->wt_radpat * o_RP_distance + o_energy;

  copy_Conf(Min, c1);

  for(iter = 0; iter < P->granularity; iter++) {
    unsigned i_atom, mol = 0;
    double Edelta, angle = 0, dir[3];
    /* loop over unique atoms. we update one at a time */
    for(i_atom = 0; i_atom < T1->n_unique + T1->Nmol; i_atom++) {
      int actual_atom;
      Tried++;
      /* the old energy is the now one if T has changed */

      /* change the configuration */
      /* make a new trial */
      renew_Trial(Delta);
      

      /* 
	 pick atom to move at random except not the zeroeth in case
	 we have three polar directions (P1 does)
      */

      actual_atom = (int) (Random() * (double) (T1->n_unique + T1->Nmol));

      /* 
	 if the space group is polar and the actual atom is the zeroeth one
	 then project off the polar component of the move 
      */

      if(G->polar && actual_atom == 0) {
	double x = Delta->dx;
	double y = Delta->dy;
	double z = Delta->dz;
	
	Delta->dx = G->polar_matrix[0][0] * x + 
	  G->polar_matrix[0][1] * y + G->polar_matrix[0][2] * z;
	
	Delta->dy = G->polar_matrix[1][0] * x + 
	  G->polar_matrix[1][1] * y + G->polar_matrix[1][2] * z;
	
	Delta->dz = G->polar_matrix[2][0] * x + 
	  G->polar_matrix[2][1] * y + G->polar_matrix[2][2] * z;
      }
      
      copy_Positions(T1,T2);
      
      if(actual_atom < T1->n_unique) {
	move_atom(T2, actual_atom, Delta, C);   
      } else {
	mol = actual_atom - T1->n_unique;   /* define mol for this
					       update */

	shift_Molecule(T2->M[mol], Delta);

	ranor(dir);
	angle = T2->M[mol]->phi * gasdev();

	rotate_Molecule(T2->M[mol], angle, dir);
      }
      
      symmetrize_Positions(T2, G);
      mod_Positions(T2);
      merge_positions(T2, P, C);
      
      clean_Grid(Gr);
      fill_Grid(Gr, T2, C);
      n_energy = energy(P, C, T2, Gr, c2);
      
      if(RP != 0) {
	get_intensity(RP, T2, C, G);
	scale_Rad_pattern(RP);
	n_RP_distance = distance_Rad_pattern(RP);
      }

      c2->RP_energy = n_RP_distance;
      c2->tot_energy = P->wt_radpat * n_RP_distance + n_energy;

      /* Metropolis test */
      Edelta = P->wt_radpat * (n_RP_distance - o_RP_distance) + 
	n_energy - o_energy;

      accepted = 1;


      if (Edelta > 0.0) {
	double a = exp(-Beta*Edelta);
	double b = Random();
	
	if(a < b) accepted = 0; /* reject criteria */

      }

      if(accepted) {
	struct Conf * ctmp;

	o_energy = n_energy;
	o_RP_distance =  n_RP_distance;

	if(c2->tot_energy < Min->tot_energy) copy_Conf(Min, c2);

	/* swap configuration data */
	ctmp = c1;
	c1 = c2;
	c2 = ctmp;

	if(actual_atom < T1->n_unique) {
	  move_atom(T1, actual_atom, Delta, C);   
	} else {
	  shift_Molecule(T1->M[mol], Delta);
	  
	  rotate_Molecule(T1->M[mol], angle, dir);
	}

	Accepted++;
      } 

      /* write the configurations to disk */
 /* MWD Change */
    mod_all_Molecules(T1->Nmol, T1->M, C);

      
#ifdef MC_ENERGY
      fprint_Conf(c1, AA); 
      fflush(AA);
      if(BB) {
 /* MWD Change */
	for(looper = 0; looper < T1->Nmol; looper++) {
	  print_Molecule(T1->M[looper], BB);
	}
	fflush(BB);
      }
#endif
    } /* loop over unique atoms */

  }

  delete_Trial(Delta);

  delete_Conf(c1);
  delete_Conf(c2);

  ret = (double) Accepted / (double) Tried;

#ifdef MC_ENERGY
  fclose(AA);
#endif

  return ret;
}
