/* 
   (C) 1997 Marco Falcioni

   radiation pattern object;
  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "error.h"
#include "constants.h"
#include "memory.h"

#include "misc.h"
#include "cell.h"
#include "verbose.h"
#include "intensity.h"
#include "spot.h"
#include "rad_pattern.h"

/*---------------------------------------------------------------------------*
 *                          Allocation of a radiation pattern                *
 *---------------------------------------------------------------------------*/

struct Rad_pattern * new_Rad_pattern(void)
{
  struct Rad_pattern * me;

  me = (struct Rad_pattern *) Malloc(sizeof(struct Rad_pattern));
  if(me == 0) {
    push_error("new_Rad_pattern",Error_Alloc_);
    exit_clean();
  }
  me->I = 0;

  me->corr = 0;
  me->composites = 0;
  me->angles = 0;

  me->narea = 0.0;
  me->xarea = 0.0;
  me->wavelength = 0.0;
  me->wavelength2 = 0.0;
  me->length = 0;
  me->type = NONE;
  me->xray_corr = 0;
  me->neu_corr = 0;
  return me;
  
}


/*---------------------------------------------------------------------------*
 *                         Destructor                                        *
 *---------------------------------------------------------------------------*/

void delete_Rad_pattern(struct Rad_pattern * me)
{
  unsigned long i;
  if(me == 0) return;
  for(i = 0; i < me->length; i ++) {
    delete_Spot(me->I[i]);
  }

  if(me->xray_corr) Free(me->xray_corr);
  if(me->neu_corr) Free(me->neu_corr);

  Free(me);
  return;
}

/*---------------------------------------------------------------------------*
 *                        Read a radiation pattern in standard format        *
 *---------------------------------------------------------------------------*/

void read_Rad_pattern(struct Rad_pattern * me, FILE * stream)
{
  long i;
  char line[1024];
  char type[1024];
  char * p;

#ifndef NOCHECKS
  if(stream == 0 || me == 0) {
    push_generic_error("read_Rad_Pattern", "(nil) args passed");
    exit_clean();
  }
#endif

  push_verbose_message("read_Rad_Pattern","reading radiation pattern");

  /* read first line: radiation type and wavelength */
  p = get_line(line, sizeof(line), stream);

  if(p == 0 || sscanf(p,"%s", type) < 1) {
    push_generic_error("read_Rad_Pattern", "error reading type or wavelength");
    exit_clean();
  }

  if(strcmp(type,"X-RAY") == 0) {
    me->type = XRAY;
    sscanf(line, "%s %lf", type, &me->wavelength);
  }
  else if(strcmp(type,"NEUTRON") == 0) {
    me->type = NEUTRON;
    sscanf(line, "%s %lf", type, &me->wavelength);
  }
  else if(strcmp(type,"BOTH") == 0) {
    me->type = BOTH;
    if(sscanf(line, "%s %lf %lf", type, &me->wavelength, &me->wavelength2)
       != 3) {
      push_generic_error("read_Rad_pattern","error reading wavelengths");
      exit_clean();
    }
  }
  else {
    push_generic_error("read_Rad_Pattern", 
		       "could not parse radiation type");
    exit_clean();
  }

  if(me->wavelength == 0.0) {
    push_generic_error("read_Rad_Pattern", 
		       "no wavelength given!");
    exit_clean();
  }

  /* read second line: number of spots */
  p = get_line(line, sizeof(line), stream);

  if(p == 0 || sscanf(line, "%d", &me->length) < 1 || me->length < 1) {
    push_generic_error("read_Rad_Pattern", 
		       "wrong number of spots read: check pxd file");
    exit_clean();
  }

  me->I = (struct Spot **) Malloc(me->length * sizeof(struct Spot *));

  if(me->I == 0) {
    push_error("read_Rad_Pattern", Error_Alloc_);
    exit_clean();
  }
  
  for(i = 0; i < me->length; i++) {
    me->I[i] = new_Spot(me->type);
    p = get_line(line, sizeof(line), stream);
    read_Spot(me->I[i], p, me->type);
    
    /* we use this area for normalization 
    if (me->type == XRAY || me->type == BOTH) 
      if(me->I[i]->xray->value > 0.0) 
	me->xarea += me->I[i]->xray->value;
    if (me->type == NEUTRON || me->type == BOTH) 
      if(me->I[i]->neutron->value > 0.0) 
      me->narea += me->I[i]->neutron->value; */
  }

  if(((me->type == XRAY || me->type == BOTH) && 
      me->I[me->length-1]->xray->value < 0.0) || 
     ((me->type == NEUTRON || me->type == BOTH) && 
      me->I[me->length-1]->neutron->value < 0.0)) {
    push_generic_error("read_Rad_pattern",
		       "the last intensity is negative: check pxd file");
    exit_clean();
  }

  /* node aliasing: negative intensities signal summation to 
     the next node intensity: the last intensity must be non negative 
     and we checked that just before...
  */
  if(me->type == XRAY || me->type == BOTH) {
    me->I[me->length - 1]->xray->composite = 
      me->I[me->length - 1]->xray;
  }
  if(me->type == NEUTRON || me->type == BOTH) {
    me->I[me->length - 1]->neutron->composite = 
      me->I[me->length - 1]->neutron;
  }

  i = 0;

  do {
    if(me->type == XRAY || me->type == BOTH) {
      if(me->I[i]->xray->value < 0.0) {
	me->I[i]->xray->composite = me->I[i+1]->xray;
      } else { 
	/* this is pretty much useless */
	me->I[i]->xray->composite = me->I[i]->xray;
      }
    } 
    if(me->type == NEUTRON || me->type == BOTH) {
      if(me->I[i]->neutron->value < 0.0) {
	me->I[i]->neutron->composite = me->I[i+1]->neutron;
      } else { 
	/* this is useless too */
	me->I[i]->neutron->composite = me->I[i]->neutron;
      }
    }
    ++i;
  } while (i < me->length);

  return;
}

/*---------------------------------------------------------------------------*
 *                Print a Radiation pattern in standard format               *
 *---------------------------------------------------------------------------*/

void print_Rad_pattern(const struct Rad_pattern * me, FILE * stream)
{
  unsigned long i;

#ifndef NOCHECKS
  if(me == 0 || stream == 0) {
    push_generic_error("print_Rad_pattern", "(nil) stream passed");
    exit_clean();
  }
#endif

  switch (me->type) {
  case XRAY:
    fprintf(stream,"X-RAY %f\n", me->wavelength);
    break;
  case NEUTRON:
    fprintf(stream,"NEUTRON %f\n", me->wavelength2);
    break;
  case BOTH:
    fprintf(stream,"BOTH %f %f\n", me->wavelength, me->wavelength2);
    break;
  case NONE:
    fprintf(stream,"NONE\n");
    break;
 }

  fprintf(stream, "%d\n", me->length);

  for(i = 0; i < me->length; i++) {
    print_Spot(me->I[i], stream);
  }
  return;
}

/*---------------------------------------------------------------------------*
 *               Print radiation pattern in format suitable for plotting     *
 *---------------------------------------------------------------------------*/
void plot_Rad_pattern(const struct Rad_pattern * me, 
		      const struct Cell * C, FILE * stream)
{
  unsigned long i;

#ifndef NOCHECKS  
  if(me == 0 || C == 0 || stream == 0) {
    push_generic_error("plot_Rad_pattern","(nil) args passed");
    exit_clean();
  }
#endif

  if(!me->angles) {
    push_generic_error("plot_Rad_pattern","call angle_Rad_pattern first");
    exit_clean();
  }

  for(i = 0; i < me->length; i++) {
    fprintf(stream, "%f ", me->I[i]->twotheta);
    print_Spot(me->I[i], stream);
  }
}

/*---------------------------------------------------------------------------*
 *        Output radiation pattern in format suitable for input              *
 *---------------------------------------------------------------------------*/

void write_Rad_pattern(const struct Rad_pattern * me, FILE * stream)
{
  unsigned long i;

#ifndef NOCHECKS  
  if(me == 0 || stream == 0) {
    push_generic_error("write_Rad_pattern","(nil) args passed");
    exit_clean();
  }
#endif

  switch (me->type) {
  case XRAY:
    fprintf(stream, "X-RAY %f\n", me->wavelength);
    break;
  case NEUTRON:
    fprintf(stream, "NEUTRON %f\n", me->wavelength);
    break;
  case BOTH:
    fprintf(stream, "BOTH %f %f\n", me->wavelength, me->wavelength2);
    break;
  case NONE:
    /* fall through */
  default:
    push_generic_error("read_Spot","weird type in switch");
    exit_clean();
  }
  
  fprintf(stream, "%d\n", me->length);

  for(i = 0; i < me->length; i++) {
    struct Spot * tmp = me->I[i] ;
    fprintf(stream, "%2d %2d %2d %2d", 
	    (int) tmp->h_, (int) tmp->k_, (int) tmp->l_,
	    tmp->multiplicity);
    if(me->type == XRAY || me->type == BOTH) 
      fprintf(stream, "  %12.7f  %d", tmp->xray->value, tmp->xray->sigma);
    
    if(me->type == NEUTRON || me->type == BOTH) 
      fprintf(stream, "  %12.7f  %d", tmp->neutron->value, tmp->neutron->sigma);
    
    fprintf(stream, "\n");
  }
}

/*---------------------------------------------------------------------------*
 *            Scale radiation pattern to minimize distance (see below)       *
 *---------------------------------------------------------------------------*/

void scale_Rad_pattern(struct Rad_pattern * RP) 
{
  unsigned long i;
  double sum_values = 0.0, sum_calcs = 0.0, factor = 0.0; 

  if( RP == 0 ) {
    return;
  }

#ifndef NOCHECKS
  if(RP->composites != 2) {
    push_generic_error("scale_Rad_pattern",
		       "call sum_composites_Rad_pattern first");
    exit_clean();
  }
#endif

  if(RP->type == XRAY || RP->type == BOTH) {
    for(i = 0 ; i < RP->length; i ++) {
      struct Intensity * sp = RP->I[i]->xray;
      if(sp->value > 0.0) {
	sum_values += sp->value * sp->calc / sp->sigma;
	sum_calcs += sp->calc * sp->calc / sp->sigma;
      }
    }
    
    factor  = sum_values/sum_calcs;
    
    for(i = 0 ; i < RP->length; i ++) 
      RP->I[i]->xray->calc *= factor;
  }

  if(RP->type == NEUTRON || RP->type == BOTH) {
    for(i = 0 ; i < RP->length; i ++) {
      struct Intensity * sp = RP->I[i]->neutron;
      if(sp->value > 0.0) {
	sum_values += sp->value * sp->calc / sp->sigma;
	sum_calcs += sp->calc * sp->calc / sp->sigma;
      }
    }
    
    factor  = sum_values/sum_calcs;
    
    for(i = 0 ; i < RP->length; i ++) 
      RP->I[i]->neutron->calc *= factor;
  }

  return;

}

/*---------------------------------------------------------------------------*
 *      Compute a distance between radiation patterns (called factor R)      *
 *---------------------------------------------------------------------------*/

double distance_Rad_pattern(struct Rad_pattern * RP) 
{
  unsigned long i;
  double res = 0.0;

  if( RP == 0 ) {
    return 0.0;
  }

#ifndef NOCHECKS
  if(RP->composites != 2) {
    push_generic_error("distance_Rad_pattern",
		       "call sum_composites_Rad_pattern first");
    exit_clean();
  }
#endif

  if(RP->type == XRAY || RP->type == BOTH) {
    for(i = 0 ; i < RP->length; ++i) {
      struct Intensity * sp = RP->I[i]->xray;
#ifdef DEBUG
      printf("spot %d v %f c %f res %f\n", i, sp->value, sp->calc, res);
#endif
      if(sp->value > 0.0)
	res += (sp->value - sp->calc) * (sp->value - sp->calc) / sp->sigma;
    }
  }

  if(RP->type == NEUTRON || RP->type == BOTH) {
    for(i = 0 ; i < RP->length; ++i ) {
      struct Intensity * sp = RP->I[i]->neutron;
      if(sp->value > 0.0)
	res += (sp->value - sp->calc) * (sp->value - sp->calc) / sp->sigma;
    }
  }

  if(isnan(res)){
    for(i = 0 ; i < RP->length; i ++) {
      struct Intensity * sp = RP->I[i]->xray;
      print_Intensity(sp, stderr);
    }
  
    exit_clean();
  }

  return 0.5 * res/(double) RP->length;
}

/*---------------------------------------------------------------------------*
 *         If some peaks are composites then sum the calculated intensities  *
 *---------------------------------------------------------------------------*/

void sum_composites_Rad_pattern(struct Rad_pattern * RP)
{
  unsigned long i;
    
  if( RP == 0 ) {
    return;
  }

#ifndef NOCHECKS
  if(RP->composites != 1) {
    push_generic_error("sum_composites_Rad_pattern",
		       "call get_intensity first");
    exit_clean();
  }
#endif

  RP->composites = 2;

  if(RP->type == XRAY || RP->type == BOTH) {
    for(i = 0 ; i < RP->length; i ++) {
      struct Intensity * sp = RP->I[i]->xray;
      if(sp->value < 0.0 && sp->composite != sp) {
	sp->composite->calc += sp->calc;
	sp->calc = 0.0;
      }
    }
  }

  if(RP->type == NEUTRON || RP->type == BOTH) {
    for(i = 0 ; i < RP->length; i ++) {
      struct Intensity * sp = RP->I[i]->neutron;
      if(sp->value < 0.0 && sp->composite != sp) {
	sp->composite->calc += sp->calc;
	sp->calc = 0.0;
      }
    }
  }

  return;
}

/*---------------------------------------------------------------------------*
 *  sort the spots: we use insertion sort                                    *
 *---------------------------------------------------------------------------*/

void sort_Rad_pattern (struct Rad_pattern * RP) 
{
  struct Spot * stmp, ** Spots;
  double amin = 180.0, dtmp;
  unsigned length, index_min = 0, i;

#ifndef NOCHECKS
  if(RP == 0) {
    push_generic_error("sort_Rad_pattern", "(nil) arg passed");
    exit_clean();
  }
#endif

  if(!RP->angles) {
    push_generic_error("sort_Rad_pattern","call angle_Rad_pattern first");
    exit_clean();
  }

  length = RP->length;
  Spots = RP->I;
  
  for(i = 0; i < length; i ++) {
    if(RP->I[i]->twotheta < amin) {
      amin = RP->I[i]->twotheta;
      index_min = i;
    }
  }

  /* here we swap the smallest element with the 0th element (sentinel) */
  stmp = Spots[0];

  Spots[0] = Spots[index_min];

  Spots[index_min] = stmp;

  /* here we use insertion sort (page 100 Sedgewick) */
  for(i = 1; i < length; ++i) {
    unsigned j = i;

    dtmp = Spots[i]->twotheta;    /* the value we want to insert */
    stmp = Spots[i];  /* along with its spot */

    while( Spots[j-1]->twotheta > dtmp ) {
      Spots[j] = Spots[j-1];    /* along with their spot */
      --j;
    }
    Spots[j] = stmp;    
  }

  return;

}

/*---------------------------------------------------------------------------*
 *  put the angles where they belong                                         *
 *---------------------------------------------------------------------------*/

void angle_Rad_pattern(struct Rad_pattern * RP, const struct Cell * C)
{
  unsigned i;

#ifndef NOCHECKS
  if(RP == 0 || C == 0) {
    push_generic_error("angle_Rad_pattern", "(nil) arg passed");
    exit_clean();
  }
#endif

  RP->angles = 1;

  for(i = 0; i < RP->length; i ++) { /* main loop over spots */
    unsigned j;
    double k[3];
    
    for(j = 0; j < 3; j ++) {
      /* this is the k vector in reciprocal space */
      k[j] = C->rec_vec[0][j] * RP->I[i]->h_ + 
	C->rec_vec[1][j] * RP->I[i]->k_ +
	C->rec_vec[2][j] * RP->I[i]->l_ ;
    }
    
    RP->I[i]->twotheta = -1.0;

    /******  if the argument is > 1 skip the spot altogether ! *****/
    if(RP->wavelength * sqrt(dot(k,k)) < 2.0) 
      RP->I[i]->twotheta = 360.0 *
	asin(0.5 * RP->wavelength * sqrt(dot(k,k))) / constPi_;

  }
  return;

} 


/*---------------------------------------------------------------------------*
 *  scale the intensity of the radiation pattern                             *
 *---------------------------------------------------------------------------*/

void norm_Rad_pattern(struct Rad_pattern * RP, enum Normalize what) 
{
  unsigned i;
  double xrmax = -100.0;
  double nmax = -100.0;

#ifndef NOCHECKS
  if(RP == 0) {
    push_generic_error("norm_Rad_pattern", "(nil) arg passed");
    exit_clean();
  }
#endif
  if(what == CALC || what == BOTH) {
    if(RP->type == XRAY || RP->type == BOTH) {
      struct Intensity * aa; 
      for(i = 0; i < RP->length; i ++) {
	aa = RP->I[i]->xray; 
	if(aa->calc > xrmax) xrmax = aa->calc;
      }  
      
      xrmax = MAX_INTENSITY/xrmax;
      
      for(i = 0; i < RP->length; i ++) {
	aa = RP->I[i]->xray; 
	if(aa->calc > 0.0) aa->calc *= xrmax;
      }  
    } 
    
    if(RP->type == NEUTRON || RP->type == BOTH) {
      struct Intensity * aa;
      
      for(i = 0; i < RP->length; i ++) {
	aa = RP->I[i]->neutron;
	if(aa->calc > nmax) nmax = aa->calc;
      }
      
      nmax = MAX_INTENSITY/nmax;
      
      for(i = 0; i < RP->length; i ++) {
	aa = RP->I[i]->neutron; 
	if(aa->calc > 0.0) aa->calc *= nmax;
      } 
    }
  } 
  
  if(what == VALUE || what == ALL) {
    if(RP->type == XRAY || RP->type == BOTH) {
      struct Intensity * aa; 
      for(i = 0; i < RP->length; i ++) {
	aa = RP->I[i]->xray; 
	if(aa->value > xrmax) xrmax = aa->value;
      }  
      
      xrmax = MAX_INTENSITY/xrmax;
      
      for(i = 0; i < RP->length; i ++) {
	aa = RP->I[i]->xray; 
	if(aa->value > 0.0) aa->value *= xrmax;
      }  
    } 
    
    if(RP->type == NEUTRON || RP->type == BOTH) {
      struct Intensity * aa;
      
      for(i = 0; i < RP->length; i ++) {
	aa = RP->I[i]->neutron;
	if(aa->value > nmax) nmax = aa->value;
      }
      
      nmax = MAX_INTENSITY/nmax;
      
      for(i = 0; i < RP->length; i ++) {
	aa = RP->I[i]->neutron; 
	if(aa->value > 0.0) aa->value *= nmax;
      } 
    }
  } 


  return;

}

/**************************************************************************

     Copy computed values to the actual values

 **************************************************************************/

void copy_Rad_pattern(struct Rad_pattern * RP)
{
  int i;

  
  for(i = 0; i < RP->length; i ++) {
    switch (RP->type) {
    case XRAY:
      if(RP->I[i]->xray->value > 0.0) RP->I[i]->xray->value = RP->I[i]->xray->calc; 
      break;
    case NEUTRON:
      if(RP->I[i]->neutron->value > 0.0) RP->I[i]->neutron->value = RP->I[i]->neutron->calc; 
      break;
    case BOTH:
      if(RP->I[i]->xray->value > 0.0) RP->I[i]->xray->value = RP->I[i]->xray->calc; 
      if(RP->I[i]->neutron->value > 0.0) RP->I[i]->neutron->value = RP->I[i]->neutron->calc;
      break;
    case NONE:
      /* fall through */
    default:
      push_generic_error("find_multiplicity",
			 "weirdness in second enum switch");
      exit_clean();
    }
  }

}


/*
void main (void) 
{
  struct Rad_pattern * p;

  connect_error_stream(stderr);

  p = new_Rad_pattern();
  read_Rad_pattern(p, stdin);
  print_Rad_pattern(p, stdout);

}
*/
