/* Created: [GJ] Fri Apr 26 01:48:41 EDT 1996
 */
#include <stdio.h>
#include <string.h>
#include "io.h"
#include "stringlib.h"
#include "resources.h"


/* Global program name resource. Separate from resource repository.    */
static char * gjprogramname;
static char gjunknownprogramname[] = "unknownProgram";


/* Global resource repository.                                         */
struct gjResourceRepository {
  struct string_list * name;
  struct string_list * value;
};
static struct gjResourceRepository gjglobalres = {
  (struct string_list *) 0,
  (struct string_list *) 0
};


/* Search repository for a named resource.                             */
static int findbyname(const char * test)
{
  int i;

  if(gjglobalres.name == (struct string_list *)0){
    return -1;
  }
  else {
    for(i=0; i<gjglobalres.name->size; i++){
      if(strcmp(test,gjglobalres.name->str[i]) == 0) return i;
    }
  }

  /* If we get here then nothing was found. */
  return -1;
}


void insert_global_resource(const char *name, const char *value)
{
  int ilook = findbyname(name);
  
  if(ilook >= 0){
    /* Found one of the same name. Replace it. */
    if(gjglobalres.value->str[ilook] != (char *)0) {
      free_string(gjglobalres.value->str[ilook]);
    }
    gjglobalres.value->str[ilook] = copy_string(value);
  }
  else {
    /* Nothing found. Add it. */
    gjglobalres.name  = append_string_to_list(gjglobalres.name, name);
    gjglobalres.value = append_string_to_list(gjglobalres.value, value);
  }
}


const char * global_resource(const char * name)
{
  int ilook = findbyname(name);
  if(ilook >= 0){
    /* Found it. */
    return gjglobalres.value->str[ilook];
  }
  else {
    /* Not found. */
    return (char *)0;
  }
}


char * extract_global_resource(const char *name)
{
  int ilook = findbyname(name);
  if(ilook >= 0){
    /* Found it. */
    char * value = copy_string(gjglobalres.value->str[ilook]);
    return value;
  }
  else {
    /* Not found. */
    return (char *)0;
  }
}


double double_from_resource(const char * key, double def)
{
  char * val;
  double result;

  if((val = extract_global_resource(key)) == 0
     || sscanf(val,"%lf",&result) != 1
     ) {
    fprintf(stderr,"%s: could not extract double from resource %s\n",
	    get_app_name(), key);
    fprintf(stderr,"%s: using default value %g\n",
	    get_app_name(), def);
    return def;
  }
  else {
    return result;
  }
}


int int_from_resource(const char * key, int def)
{
  char * val;
  int result;

  if((val = extract_global_resource(key)) == 0
     || sscanf(val,"%d",&result) != 1
     ) {
    fprintf(stderr,"%s: could not extract int from resource %s\n",
	    get_app_name(), key);
    fprintf(stderr,"%s: using default value %d\n",
	    get_app_name(), def);
    return def;
  }
  else {
    return result;
  }
}


int bool_from_resource(const char * key, int def)
{
  char * val;

  if((val = extract_global_resource(key)) == 0) {
    fprintf(stderr,"%s: could not extract int from resource %s\n",
	    get_app_name(), key);
    fprintf(stderr,"%s: using default value %d\n",
	    get_app_name(), def);
    return def;
  }
  else {
    if(   strcmp(val,"on")==0
       || strcmp(val,"true")==0
       || strcmp(val,"yes")==0
      ){
      return 1;
    }
    else if(   strcmp(val,"off")==0
            || strcmp(val,"false")==0
            || strcmp(val,"no")==0
      ){
      return 0;
    }
    else {
      fprintf(stderr,"%s: bad val %s for (bool) resource %s: using \"off\"\n",
	      get_app_name(), val, key);
      return 0;
    }
  }
}


void set_app_name(const char * str)
{
  gjprogramname = copy_string(str);
}


const char * get_app_name(void)
{
  if(gjprogramname == (char *)0){
    return gjunknownprogramname;
  }
  else {
    return gjprogramname;
  }
}


int args_to_resources(int argc, char * argv[],
		      const struct string_list * checklist)
{
  int flag = 0;

  int i;
  int arglen;
  char * eqsign;
  set_app_name(argv[0]);
  
  for(i=1; i<argc; i++){

    /* Check for name=value form. */
    eqsign = strchr(argv[i], '=');

    if(eqsign == (char *)0){
      /* No name, just a value. */

      /* Note mode and discard leading "-/+" sign. */
      char mode[5];
      if(argv[i][0] == '-'){
	argv[i]++;
	strcpy(mode, "off");
      }
      else if(argv[i][0] == '+'){
	argv[i]++;
	strcpy(mode, "on");
      }
      else {
	strcpy(mode, "on");
      }

      insert_global_resource(argv[i], mode);
    }
    else {
      /* name=value pair */
      
      int valuelen;
      int namelen;

      /* Do not keep a leading "-" sign. */
      if(argv[i][0] == '-'){
	argv[i]++;
      }
      arglen = strlen(argv[i]) + 1;     /* includes trailing \0 */

      valuelen = strlen(eqsign+1) + 1;  /* includes trailing \0 */
      namelen = arglen - valuelen -1;   /* includes space for \0 because of
					   the "=" sign */
      if(namelen <= 0 ){
	fprintf(stderr,
		"%s: args_to_resources: badly formed option %s\n",
		argv[0], argv[i]);
	flag = 1;
      }
      else {
	char * name = copy_string(argv[i]);
	name[namelen] = '\0';
	if(checklist == 0 || in_string_list(checklist, name)){
	  insert_global_resource(name, eqsign+1);  /* eqsign+1=ptr to val */
	}
	else {
	  fprintf(stderr,"%s: unrecognized option %s\n", argv[0], argv[i]);
	  flag = 1;
	}
	
	free_string(name);
      }
    }
  }
  
  return flag;
}


int scan_resources(FILE *fp)
{
  int flag = 0;
  char * line;

  if(fp != (FILE *)0){
    while((line = fgetnextline(fp)) != (char *)0){
      if(!emptyline(line,"#%")){
	/* Not the most efficient way. But safe. */
	char * newline = copy_string(line);
	char * buff1;
	char * buff2;
	charsub_string(newline, ":", ' ');
	buff1 = new_string(strlen(newline)+1);
	buff2 = new_string(strlen(newline)+1);
	if(sscanf(newline,"%s%s", buff1, buff2) != 2){
	  fprintf(stderr,"%s: scan_resources: bad resource line:\n",
		  get_app_name());
	  fprintf(stderr,"%s\n", line);
	  flag = 1;
	}
	else {
	  insert_global_resource(buff1, buff2);
	}
	free_string(buff1);
	free_string(buff2);
	free_string(newline);
      }
      free_string(line);
    }
  }
  else {
    fprintf(stderr,"%s: scan_resources: cannot scan from (nil)\n",
	    get_app_name());
    flag = 1;
  }

  return flag;
}


int dump_resources(FILE *fp, const char * prefix)
{
  int flag = 0;
  int i;

  if(fp != (FILE *)0){
    for(i=0; i<gjglobalres.name->size && i<gjglobalres.value->size; i++){
      fprintf(fp,"%s%s: %s\n",
	      prefix,
	      gjglobalres.name->str[i],
	      gjglobalres.value->str[i]
	      );
    }
  }
  else {
    fprintf(stderr,"%s: dump_resources: cannot dump to (nil)\n",
	    get_app_name());
    flag = 1;
  }

  return flag;
}
