/* Created: [GJ] Fri Apr 26 04:46:13 EDT 1996
 */
#ifndef GJSTRINGS_H
#define GJSTRINGS_H


struct string_list {
  int size;
  char ** str;
};


/* Allocate a (char *) string.
 * Returns (char *)0 on failure.
 */
extern char * new_string(unsigned long size);


/* Free a (char *) string. */
extern void free_string(char * s);


/* Create a string value and copy a string into it.
 * Basically a value-checked strcpy().
 * The argument is not destroyed.
 */
extern char * copy_string(const char * s);


/* Check if a character is a member of a given string. */
extern int in_string(int ch, const char * str);


/* Append a character to the end of a string, reallocating.
 * The argument is destroyed. Obviously not very efficient,
 * use in moderation.
 */
extern char * append_char_string(char * str, char c);


/* Mangle a string literal. Maps
 *                           ' ' --> '\a'
 *                           '\n' --> '\b'
 */
extern void mangle_string(char * s);


/* Unmangle a string literal. Not exactly the reverse of mangle, since
 * it turns \a and \b back into spaces ' '. Therefore \n are purged.
 */
extern void unmangle_string(char * s);


/* Replace occurences of certain characters with another character.
 * Any character from list (char *)badlist is replaced by (char) ch.
 * As an example, this could be used to replace various white space
 * characters with blanks, a typical normalization task.
 * The string is transformed in place.
 */
extern void charsub_string(char * str, const char * badlist, char ch);


/* Concatenate strings. A value-checked version of strcat().
 * The arguments are not destroyed.
 * Note that a call of the form
 *    str = concat_strings(str, str2)
 * does NOT make sense; it will produce a memory leak. Instead use
 * the function append_string() below.
 * The returned string should be freed by the client using free_string().
 */
extern char * concat_strings(char * s1, const char * s2);


/* Append a string to another. This automatically destroys the
 * first argument so that the call
 *     str = append_string(str, str2);
 * makes sense.
 * The returned string can be freed by the client using free_string().
 */
extern char * append_string(char * str1, const char * str2);


/* Allocate a string_list.
 * It is initialized to zero size.
 */
extern struct string_list * new_string_list(void);


/* Free a string_list. */
extern void free_string_list(struct string_list *);


/* Create a string_list from a (char **) of a given size.
 * The elements of the original are duplicated so that it can be
 * destroyed at a later point without side effects.
 */
extern struct string_list * create_string_list_from_array(const char **, int);


/* Split a string into fields separated by separator characters. */
extern struct string_list * split_string(const char * str,
					 const char * separators);


/* Copy a string list.
 * Duplicates contents.
 * The argument is not destroyed.
 */
extern struct string_list * copy_string_list(const struct string_list *);


/* Given an array of strings, increase its size by one
 * and add a given string. The string is duplicated.
 * The original string list is destroyed.
 */
extern struct string_list * append_string_to_list(struct string_list *,
						  const char *);


/* Concatenate string lists.
 * The arguments are not destroyed.
 */
extern struct string_list *concat_string_lists(const struct string_list *,
					       const struct string_list *);


/* Test if a string is in a given list.                   */
extern int in_string_list(const struct string_list *, const char *);


/* Scan a double value from a string with error checking. */
extern double double_from_string(const char *);


/* Scan an int value from a string with error checking.   */
extern int int_from_string(const char *);



#endif /* GJSTRINGS_H */
