/*

  Marco Falcioni (C) 1997

  here we set up the space group and maybe we get a naming convention also
  a routine to test the goodness of the subgroups specified is provided

  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "memory.h"
#include "matrices.h"
#include "error.h"
#include "group.h"
#include "symmetry.h"
#include "misc.h"
#include "verbose.h"


#ifndef INLINE
#define INLINE
#endif

INLINE double * norm_vec(double *);
int eig_no_good(struct Sym_group * G, double *);

void delete_Sym_group (struct Sym_group * G)
{
  unsigned long i;
  if(G == 0) return;
  for(i = 0; i < G->n_el; i++) {
    delete_Group_el(G->list[i]);
  }
  return;
}


/* 
   This routine tests that Sym_group is actually a group
   a group is defined by elements, a group multiplication rule;
   also, it must be associative and closed under the rule. Every element 
   must possess an inverse and the identity must be present
*/


void test_Sym_group (struct Sym_group * G)
{
  /* these are needed to ensure we recognize products and sums */
  double eps = 1.0e-3; 
  long i,j,k,n,m;
  char flag = 0;
  long ** table;
  long Identity = 0;
  struct Group_el * Temp = new_Group_el();

  table = new_matrix_l(G->n_el);

  /* find identity */
  for(k=0; k< G->n_el; k++) {
    double sum = 0.0;
    for(i=0; i<3; i++) {
      for(j=0; j<3; j++) {
	if(i==j)
	  sum += fabs(G->list[k]->W[i][j]-1.0);
	else
	  sum += fabs(G->list[k]->W[i][j]);
      }
    }
    if(sum < eps) {
      flag = 1;
      Identity = k;
      break;
    }
  }

  if(flag == 0) {
    push_generic_error("test_Sym_group","identity not found in group!");
    exit_clean();
  }
  flag = 0;

  /* put identity as first list member if not yet there */
  if(Identity != 0) {
    copy_Group_el(Temp, G->list[0]);
    copy_Group_el(G->list[Identity], G->list[0]);
    copy_Group_el(Temp, G->list[Identity]);
  }

  Identity = 0;

  /* closure test */

  for(k=0; k< G->n_el; k++) {
    for (n=0; n < G->n_el;  n ++) {  /* we need all possible products since   
					elements do not commute! */
      double sum = 0.0;


      flag = 0;
      table[k][n] = -1;

      mult_Group_el(Temp, G->list[k], G->list[n]);
#ifdef DEBUG
      write_Group_el(Temp, stdout);
      printf("closure: element %d and %d\n",k,n);
#endif
      /* compare Temp with all group elements */
      for(m=0; m < G->n_el; m++) {
	sum = comp_Group_el(Temp, G->list[m]);
#ifdef DEBUG
	  printf("element %d and %e %e\n",m, sum, eps);
#endif
	if (sum < eps) {/* we found the element */
	  table[k][n] = m;
	  flag = 1;
	  break;
	}
      }
      if(flag == 0) {
	push_generic_error("test_Sym_group", "the group is not closed!");
	exit_clean();
      }
    }
  }

  /*  check the exsistence of an inverse */
  
  for(k=0; k< G->n_el; k++) {
    flag = 0;
    for (n=0; n < G->n_el;  n ++) 
      if( table[k][n] == Identity && table[n][k] == Identity ) {
	flag = 1; /* the inverse is in the multiplication table */
	break;
      }

    if(flag == 0) {
      char buff[1024];
	sprintf(buff,"element %ld has no inverse",k);
	push_generic_error("test_symm_Group",buff);
	exit_clean();
    } 
  }
  flag = 0;

#ifdef DEBUG
  for(k=0; k< G->n_el; k++) {
    for (n=0; n < G->n_el; n++) {
      printf("%d ", table[k][n]);
    }
    printf("\n");
  }
#endif

  /* checking associativity, from the multiplication table */


  for(k=0; k< G->n_el; k++)
    for (n=0; n < G->n_el; n++)
      for (m=0; m < G->n_el; m++)
	if(table[table[k][n]][m] != table[k][table[n][m]]){
	  char buff[1024];
	  sprintf(buff,"product %ld %ld %ld not associative", 
		  k,n,m);
	  push_generic_error("test_Sym_group",buff);
	  flag = 1;
	}
  if(flag == 1) {
    push_generic_error("test_Sym_group", 
		       "The symmetry operations do not represent a group");
    exit_clean();
  } 

  G->test = 1; /* test passed */

  delete_Group_el(Temp);
  return;

}

void find_polar_Sym_group(struct Sym_group * G) 
{
  unsigned long i;
  double A[3][3];
  double tiny = 1E-15;
  char is_zero = 1;

  double * e1 = 0, * e2 = 0;
  double norm;
    
  double eig1[3] = {0.0, 0.0, 0.0}; 
  double eig2[3] = {0.0, 0.0, 0.0};
  double eig3[3] = {0.0, 0.0, 0.0};

  int no_eig1 = 0, no_eig2 = 0, no_eig3 = 0, eigs;
  double n1, n2, denominator;
  double ** matrix;

  if(G == 0) {
    push_generic_error("find_polar_Sym_group","(nil) stream passed");
    exit_clean();
  }

  G->polar = 0;
  G->polar_matrix = 0;
  G->move_0 = 1;

  push_verbose_message("find_polar_Sym_group", 
		       "looking for common eigenvalues in the symmetry group");

  /* if the group is centrosymmetric it is nonpolar */
  if(is_csym_Sym_group(G)) {
    push_verbose_message("find_polar_Sym_group", 
			 "no polar directions found");
    return;
  }

  /* zero the matrix A */
  for(i = 0; i < 3; ++i) {
    unsigned j;
    for(j = 0; j < 3; ++j) {
      A[i][j] = 0.0;
    }
  }

  /* A is the sum of all the group elements */

  for(i = 0; i < G->n_el; ++i) {
    unsigned j;
    for(j = 0; j < 3; ++j) {
      unsigned k;
      for(k = 0; k < 3; ++k) {
	A[j][k] += G->list[i]->W[j][k];	
      }
    }
  }
  
  for(i = 0; i < 3; ++i) {
    unsigned j;
    for(j = 0; j < 3; ++j) {
      A[i][j] /= (double) G->n_el;
      if(A[i][j] > tiny) is_zero = 0;
    }
  }
  
  if(is_zero) {
    push_verbose_message("find_polar_Sym_group", 
			 "no polar directions!");
    return;
  }
  
  /* find the eigenvalues of the matrix with eigenvalue 1 */
  
  /* first attempt x3 = 1.0 */
 
  eig1[0] = 1.0;
  eig1[1] = 0.0;
  eig1[2] = 0.0;
  
  denominator = -1.0 + A[1][1] + A[1][2] * A[2][1] + A[2][2] - A[1][1] * A[2][2];
  
  n1 = - (A[2][0] + A[1][1] * A[2][0] - A[1][0] * A[2][1]);
  n2 = - (A[1][0] + A[1][2] * A[2][0] - A[1][0] * A[2][2]);
  
  if(denominator > tiny) { /* this is all fine: the eigenvalue exists */
    eig1[1] = n1/denominator;
    eig1[2] = n2/denominator;
  } else {
    if(n1 > tiny || n2 > tiny || fabs(A[0][0] - 1.0) > tiny) { 
      /* the denominator is zero and at least 
				    one numerator is nonzero: no eigenvalue */
      no_eig1 = 1;
    }
  }
  
  /* second attempt x2 = 1.0 */
  eig2[0] = 0.0;
  eig2[1] = 1.0;
  eig2[2] = 0.0;
  
  denominator = -1 + A[0][0] + A[0][2]* A[2][0] + A[2][2] - A[0][0] * A[2][2];
  n1 = -( A[0][1] + A[0][2] * A[2][1] - A[0][1] * A[2][2]);
  n2 = -( A[0][1] * A[2][0] + A[2][1] - A[0][0] * A[2][1]);
  
  if(denominator > tiny) { /* this is all fine: the eigenvalue exists */
    eig2[0] = n1/denominator;
    eig2[2] = n2/denominator;
  } else {
    if(n1 > tiny || n2 > tiny || fabs(A[1][1] - 1.0) > tiny) { 
      /* the denominator is zero and at least 
	 one numerator is nonzero: no eigenvalue */
      no_eig2 = 1;
    } 
  }
   
  /* third case x2 = 1 */
  eig3[0] = 0.0;
  eig3[1] = 0.0;
  eig3[2] = 1.0; /* this is an arbitrary constraint */
  
  denominator = -1.0 + A[0][0] + A[0][1] * A[1][0] + A[1][1] - A[0][0] * A[1][1];
  n1 = - (A[0][2]  -  A[0][2] * A[1][1] + A[0][1] * A[1][2]);
  n2 = - (A[0][2] * A[1][0] + A[1][2] - A[0][0] * A[1][2]);
  
  if(denominator > tiny) { /* this is all fine: the eigenvalue exists */
    eig3[0] = n1/denominator;
    eig3[1] = n2/denominator;
  } else {
    if(n1 > tiny || n2 > tiny || fabs(A[2][2] - 1.0) > tiny) { 
      /* the denominator is zero and at least 
	 one numerator is nonzero, or the diagonal element is not 1: no eigenvalue */
      no_eig3 = 1;
    } 
  }
 
  /* check that the eigenvalues found are OK */
 
  if(!no_eig1 && eig_no_good(G, eig1)) {
    no_eig1 = 1;
  }

  if(!no_eig2 && eig_no_good(G, eig2)) {
    no_eig2 = 1;
  }

  if(!no_eig3 && eig_no_good(G, eig3)) {
    no_eig3 = 1;
  }

  eigs = 3 - no_eig1 - no_eig2 - no_eig3;

  if(eigs == 0) {  /* no polar directions in group */
    push_verbose_message("find_polar_Sym_group", 
			 "no polar directions found");
    return;
  } else if (eigs == 3) {
    /* special case */
    G->polar = 1;
    G->move_0 = 0;
    push_verbose_message("find_polar_Sym_group", 
			 "three polar directions found");
    return;
  }

  matrix = new_matrix_d(3);

  /* make the matrix the identity */
  
  for(i = 0; i < 3; ++i) {
    unsigned long j;
    for(j = 0; j < 3; ++j) {
      if(i == j) matrix[i][j] = 1.0;
      else matrix[i][j] = 0.0;
    }
  }
  
  /* 2 cases are left: 1 or 2 vectors 
     we only need to make them orthonormal if we have 2 */
  
  switch (eigs) {
  case 1: 
    /* get the eigenvalue */
    if(!no_eig1) e1 = eig1;
    if(!no_eig2) e1 = eig2;
    if(!no_eig3) e1 = eig3;
    
    /* normalize the eigenvalue */

    e1 = norm_vec(e1);
    
    /* construct projector P = 1 - e ^ e (^ is nonal form) */
    
    for(i = 0; i < 3; ++i) {
      unsigned long j;
      for(j = 0; j < 3; ++j) {
	matrix[i][j] -= e1[i] * e1[j];
      }
    }
    
    G->polar = 1;
    G->polar_matrix = matrix;
    G->move_0 = 1;

    push_verbose_message("find_polar_Sym_group", 
		       "one polar direction found");
    break;

  case 2: /* we have two eigenvlaues */
    if(!no_eig1 && !no_eig2) {
      e1 = eig1;
      e2 = eig2;
    } else if (!no_eig1 && !no_eig3) {
      e1 = eig1;
      e2 = eig3;
    } else if (!no_eig2 && !no_eig3) {
      e1 = eig2;
      e2 = eig3;
    }
    
    /* make vectors orthogonal */
    norm = dot(e1, e2);

    for(i = 0; i < 3; ++i) {
      e2[i] -= norm * e1[i];
    }

    /* normalize vectors */

    e1 = norm_vec(e1);
    e2 = norm_vec(e2);
   
    /* construct projector P = 1 - e1 ^ e1 - e2 ^ e2 (^ is nonal form) */

    for(i = 0; i < 3; ++i) {
      unsigned long j;
      for(j = 0; j < 3; ++j) {
	matrix[i][j] -= e1[i] * e1[j] + e2[i] * e2[j];
      }
    }

    G->polar = 1;
    G->polar_matrix = matrix;
    G->move_0 = 1;

    push_verbose_message("find_polar_Sym_group", 
		       "two polar directions found");
    break;

  default:
    push_generic_error("find_polar_Sym_group","error in switch");
    exit_clean();
    
  }
  
  return;
}

int eig_no_good(struct Sym_group * G, double * eig1)
{
  unsigned long i;
  double tiny = 1e-6;

  for(i = 0; i < G->n_el; ++i) {
    double ** mat = G->list[i]->W;
    double vec[3], mod;
    unsigned a;
    for(a = 0; a < 3; ++a) {
      unsigned b;
      for(b = 0; b < 3; ++b) {
	vec[a] = mat[a][b] * eig1[b];
      }
    }
    mod = dot(vec,vec);

    if(abs(mod - 1.0) < tiny) return 1; /* one is enough */

  }
  return 0;
}

INLINE 
double * norm_vec(double * vec)
{
  unsigned i;
  double norm;
  static double tiny = 1.0e-10;

#ifndef NOCHECKS
  if(vec == 0) {
    push_generic_error("norm_vec","(nil) arg passed");
    exit_clean();
  }
#endif

  norm = dot(vec,vec);

  if(norm < tiny) {
    push_generic_error("norm_vec","very small vector to normalize");
    exit_clean();
  }
  
  for(i = 0; i < 3; ++i) vec[i] /= norm;

  return vec;

}

/* 
   given the number of group elements, it reads them all in from
   stream and returns an allocated structure containing the whole
   symmetry group.  
*/

struct Sym_group * read_Sym_group (unsigned long n_el, FILE * stream)
{
  unsigned long i;
  struct Sym_group * G;

  if(stream == 0) {
    push_generic_error("read_Sym_group","(nil) stream passed");
    exit_clean();
  }

  push_verbose_message("read_Sym_group", "reading symmetry operators");

  G = (struct Sym_group *) Malloc(sizeof(struct Sym_group));

  G->number = 0;
  strcpy(G->label, "None");
  strcpy(G->qualifier, "None");

  G->n_el = n_el;
  G->list = (struct Group_el **) Malloc(n_el * sizeof(struct Group_el *));

  for(i=0; i<n_el; i++)
    {
      G->list[i] = read_Group_el(stream);
    }
  
  G->test = 0; /* the group was not tested */

  return G;
}

void write_Sym_group (const struct Sym_group * G, FILE * stream)
{
  unsigned long i;
  
  if(stream == 0 || G == 0) {
    push_generic_error("write_Sym_group","(nil) args passed");
  }

  fprintf(stream,"# SYMMETRY  NUMBER %d  LABEL %s  QUALIFIER %s\n",
	  G->number, G->label, G->qualifier);
  
  for(i=0; i < G->n_el; i++)
    write_Group_el(G->list[i], stream);
  
  return;
}


struct Sym_group * read_Sym_group_XTL (unsigned long n_el, FILE * stream)
{
  unsigned long i, j;
  struct Sym_group * G;
  char line[1024], * p;
  char a[7][20] = {"", "", "", "", "", "", ""};


  if(stream == 0) {
    push_generic_error("read_Sym_group_XTL","(nil) stream passed");
    exit_clean();
  }

  push_verbose_message("read_Sym_group_XTL", 
		       "reading symmetry operators");

  G = (struct Sym_group *) Malloc(sizeof(struct Sym_group));

  p = get_line(line, sizeof(line), stream);

  i = sscanf(p, "%s %s %s %s %s %s %s", a[0], a[1], a[2], a[3], a[4], a[5], a[6]);

  if(strcmp(a[0], "SYMMETRY")) { /* 0 is ture in this case */
    push_generic_error("read_Sym_group_XTL",
		       "cannot read symmetry group 1st line A");
    exit_clean();
  }

  G->number = 0;
  strcpy(G->label, "None");
  strcpy(G->qualifier, "None");

  for(j=1; j < i; j += 2) {
    if(!strncmp(a[j],"NUMBER", (size_t) 6)) {
      G->number = (unsigned) strtol(a[j+1], /* rtn ptr NULL */ 0, /* base */ 10);
    } else if (!strncmp(a[j],"LABEL", (size_t) 6)) {
      strcpy(G->label, a[j+1]);
    } else if (!strncmp(a[j],"QUALIFIER", (size_t) 9)) {
      strcpy(G->qualifier, a[j+1]);
    } else {
       push_generic_error("read_Sym_group_XTL","cannot read symmetry group 1st line B");
       exit_clean();
    }
  }

  G->n_el = n_el;
  G->list = (struct Group_el **) Malloc(n_el * sizeof(struct Group_el *));

  for(i=0; i<n_el; i++)
    {
      G->list[i] = read_Group_el_XTL(stream);
    }
  
  G->test = 0; /* the group was not tested */

  return G;
}


void write_Sym_group_XTL(const struct Sym_group * G, FILE * stream)
{
  unsigned long i;
  if(stream == 0 || G == 0) {
    push_generic_error("write_Sym_group_XTL","(nil) args passed");
  }

  fprintf(stream,"SYMMETRY");
  
  if(G->number > 0) fprintf(stream, " NUMBER %d", G->number);
  if(strcmp(G->label, "None")) fprintf(stream, " LABEL %s", G->label);
  if(strcmp(G->qualifier, "None")) 
    fprintf(stream, " QUALIFIER %s", G->qualifier);

  fprintf(stream, "\n");

  for(i=0; i < G->n_el; i++)
    write_Group_el_XTL(G->list[i], stream);
  
  return;

}


/*****************************************************************************\
 *                                                                           *
 *             Check that -Identity is a group element                       *
 *                                                                           *
\*****************************************************************************/
int is_csym_Sym_group(const struct Sym_group * G)
{
  double eps = 1e-6;
  int k, flag = 0;
  /* find -identity */

  for(k = 0; k < G->n_el; k++) {
    struct Group_el * aa = G->list[k];
    int i;
    double sum = 0.0;
    for(i = 0; i < 3; i++) {
      int j;

      sum += fabs(aa->w[i]);  /* translations count */

      for(j = 0; j < 3; j++) {
	if(i == j)
	  sum += fabs(aa->W[i][j]+1.0);
	else
	  sum += fabs(aa->W[i][j]);
      }
    }
    if(sum < eps) {
      flag = 1;
      break;
    }
  }
  
  return flag;

}
