/*

  Sticky Balls 0.01
  Marco Falcioni (C) 1999

  (C) 1999 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "error.h"
#include "misc.h"

#include "vector_functions.h"

void cross_product(const double * a, const double * b, double * c)
{

#ifndef NOCHECKS
  if(a == 0 || b == 0 || c == 0) {
    push_generic_error("cross_product", "nil args passed");
    exit_clean();
  }
#endif

  c[0] =    a[1] * b[2] - a[2] * b[1];
  c[1] =  - a[0] * b[2] + a[2] * b[0];
  c[2] =    a[0] * b[1] - a[1] * b[0];

}


void scale_vector(double * a, double b, int size)
{
  int i;
#ifndef NOCHECKS
  if(a == 0 || b == 0.0 || size < 1) {
    push_generic_error("scale_vector", "nil args passed");
    exit_clean();
  }
#endif
  for(i = 0; i < size; i++){
    a[i] *= b;
  }

}

void add_vectors(const double * a, const double * b, double * c, 
		 double m, int size)
{
  int i;
#ifndef NOCHECKS
  if(a == 0 || b == 0 || c == 0 || size < 1) {
    push_generic_error("add_vectors", "nil args passed");
    exit_clean();
  }
#endif
  for(i = 0; i < size; i++) c[i] = a[i] + m * b[i];

}

double norm2_vector(const double * a, int size) 
{
  double n = 0.0;
  int i;

#ifndef NOCHECKS
  if(a == 0) {
    push_generic_error("norm2_vector", "nil args passed");
    exit_clean();
  }
#endif
  
  for(i = 0; i < size; i ++) n += a[i] * a[i];

  return n;

}


/* from Frenkel/Smit Page 410 */

void ranor( double * a ) 
{
  double ransq, ran1, ran2, ranh;

#ifndef NOCHECKS
  if(a == 0) {
    push_generic_error("ranor", "nil args passed");
    exit_clean();
  }
#endif

  do {
    ran1 = 1.0 - 2.0 * Random();
    ran2 = 1.0 - 2.0 * Random();
    ransq = ran1 * ran1 + ran2 * ran2;
  }  while(ransq > 1.0);

  ranh = 2.0 * sqrt(1.0 - ransq);

  a[0] = ran1 * ranh;
  a[1] = ran2 * ranh;
  a[2] = 1.0 - 2.0 * ransq;

}

double * copy_vector(double * a, const double * b, int size)
{
  int i;
#ifndef NOCHECKS
  if(a == 0 || b == 0 || size < 1) {
    push_generic_error("copy_vector", "nil args passed");
    exit_clean();
  }
#endif
  
  for( i = 0; i < size; i++) a[i] = b[i];

  return a;
}

/* This routine rotates the vector a about n by theta 
   (counterclockwise is +) 
  r' = a cos(theta) + n(n.a)(1-cos(theta)) + n x a sin(theta)  */

void rotate_vector(const double * a, const double * n, double * ret, 
		double cost, double sint)  {
  double fact; 
  double v[3]; 

#ifndef NOCHECKS
  if(a == 0 || n == 0 || ret == 0) {
    push_generic_error("copy_vector", "nil args passed");
    exit_clean();
  }
#endif

  fact = (n[0] * a[0] + n[1] * a[1] + n[2] * a[2]) * (1. - cost); 

  v[0] =   n[1] * a[2] - n[2] * a[1];
  v[1] = - n[0] * a[2] + n[2] * a[0];
  v[2] =   n[0] * a[1] - n[1] * a[0];

  ret[0] = a[0] * cost + n[0] * fact + v[0] * sint; 
  ret[1] = a[1] * cost + n[1] * fact + v[1] * sint; 
  ret[2] = a[2] * cost + n[2] * fact + v[2] * sint; 
  return;  

}





